#ifndef uint32_t
#define uint32_t unsigned int
#endif

#define H0 0x6a09e667
#define H1 0xbb67ae85
#define H2 0x3c6ef372
#define H3 0xa54ff53a
#define H4 0x510e527f
#define H5 0x9b05688c
#define H6 0x1f83d9ab
#define H7 0x5be0cd19


uint rotr(uint x, int n) {
  if (n < 32) return (x >> n) | (x << (32 - n));
  return x;
}

uint ch(uint x, uint y, uint z) {
  return (x & y) ^ (~x & z);
}

uint maj(uint x, uint y, uint z) {
  return (x & y) ^ (x & z) ^ (y & z);
}

uint sigma0(uint x) {
  return rotr(x, 2) ^ rotr(x, 13) ^ rotr(x, 22);
}

uint sigma1(uint x) {
  return rotr(x, 6) ^ rotr(x, 11) ^ rotr(x, 25);
}

uint gamma0(uint x) {
  return rotr(x, 7) ^ rotr(x, 18) ^ (x >> 3);
}

uint gamma1(uint x) {
  return rotr(x, 17) ^ rotr(x, 19) ^ (x >> 10);
}

uint inversemod16(uint x, uint y) {
	return (x+16-y)%16;
}


void _sha256(uchar *message, uint mlen, uchar *digest){
	uint total = ((mlen%64>=56)?2:1)+mlen/64;
	uint workspace[8],W[16], temp, A,B,C,D,E,F,G,H,T1,T2;
	// TODO: optimize endianess and remove loops

	//Padding the message
	uint padding_iter = mlen;
	message[padding_iter++] = 0x80;
	while((padding_iter+8)%64) message[padding_iter++] = 0;
	for(uint iter=0,rlen=(mlen<<3); iter<8; iter++) {
		message[padding_iter+8-1-iter] = (unsigned char) (rlen & 0xff);
		rlen >>= 8;
	}

	uint K[64]={
0x428a2f98, 0x71374491, 0xb5c0fbcf, 0xe9b5dba5, 0x3956c25b, 0x59f111f1, 0x923f82a4, 0xab1c5ed5, 
0xd807aa98, 0x12835b01, 0x243185be, 0x550c7dc3, 0x72be5d74, 0x80deb1fe, 0x9bdc06a7, 0xc19bf174, 
0xe49b69c1, 0xefbe4786, 0x0fc19dc6, 0x240ca1cc, 0x2de92c6f, 0x4a7484aa, 0x5cb0a9dc, 0x76f988da, 
0x983e5152, 0xa831c66d, 0xb00327c8, 0xbf597fc7, 0xc6e00bf3, 0xd5a79147, 0x06ca6351, 0x14292967,
0x27b70a85, 0x2e1b2138, 0x4d2c6dfc, 0x53380d13, 0x650a7354, 0x766a0abb, 0x81c2c92e, 0x92722c85,
0xa2bfe8a1, 0xa81a664b, 0xc24b8b70, 0xc76c51a3, 0xd192e819, 0xd6990624, 0xf40e3585, 0x106aa070,
0x19a4c116, 0x1e376c08, 0x2748774c, 0x34b0bcb5, 0x391c0cb3, 0x4ed8aa4a, 0x5b9cca4f, 0x682e6ff3,
0x748f82ee, 0x78a5636f, 0x84c87814, 0x8cc70208, 0x90befffa, 0xa4506ceb, 0xbef9a3f7, 0xc67178f2
};

  workspace[0] = H0;
  workspace[1] = H1;
  workspace[2] = H2;
  workspace[3] = H3;
  workspace[4] = H4;
  workspace[5] = H5;
  workspace[6] = H6;
  workspace[7] = H7;
  for(uint item=0; item<total; item++) {
	
	uint *plaintext = (void *) message;
	uint *msgchunk = plaintext+(item<<4);

    A = workspace[0];
    B = workspace[1];
    C = workspace[2];
    D = workspace[3];
    E = workspace[4];
    F = workspace[5];
    G = workspace[6];
    H = workspace[7];

	uint flag = 0;
#define HASHCAL \
		W[t%16] = w; \
		uint T1 = H+sigma1(E)+ch(E,F,G)+K[t]+w; \
		uint T2 = sigma0(A)+maj(A,B,C); \
		H=G; G=F;F=E; E=D+T1; D=C; C=B; B=A; A=T1+T2; 
	for(uint t=0; t<16; t++) {
		uint w = 0;
		for(uint endian_idx=0; endian_idx<4; endian_idx++) {
			uint temp = (msgchunk[t]>>(endian_idx<<3))&0xff;
			w |= (temp<<((3-endian_idx)<<3));
		}
		HASHCAL
	}
	for(uint t=16; t<64; t++) {
		uint w = gamma1(W[inversemod16(t,2)])+W[inversemod16(t,7)]
				+gamma0(W[inversemod16(t,15)])+W[inversemod16(t,16)];
		HASHCAL
	}
#undef HASHCAL
    workspace[0] += A;
    workspace[1] += B;
    workspace[2] += C;
    workspace[3] += D;
    workspace[4] += E;
    workspace[5] += F;
    workspace[6] += G;
    workspace[7] += H;
	}
	for(uint iter=0; iter<32; iter++) {
		uint quoten = iter & 0xfffffffc;
		uint remain = iter & 0x00000003;
		digest[iter] = ((unsigned char *) ((void *) workspace))[quoten+(3-remain)];
	}
}

__kernel void sha256( __constant uchar *message, __constant uchar*mlen, __global uchar *digest) {
	unsigned int mlen_copy = *mlen;
	unsigned char message_copy[4096],digest_buf[32];
	for(uint i=0; i<mlen_copy; i++) {
		message_copy[i]=message[i];
	}
	_sha256(message_copy,mlen_copy,digest_buf);
	for(uint i=0; i<32; i++) {
		digest[i]=digest_buf[i];
	}
}
